
from urllib.parse import quote

from .auxiliaries import as_integer, apply_aspect_ratio, compute_focal_alignment, format_styles, \
    apply_fit_and_alignment_styles

from csskit.common import CSSUnits, CSSUnitValue

__all__ = ['render_content_in_responsive_cell']


def default_renderer(cell, content, dimensions, adjustments, alignment, attributes, layout_sizes):
    return "<div></div>"


def render_content_in_responsive_cell(content, intrinsic_dimensions, focal_point, cell, renderer, attributes=None,
                                      is_replaced_element=True):

    # Use the defined cell aspect ratio or the intrinsic aspect ratio of the content
    cell_aspect_ratio = cell.aspect_ratio if cell.use_aspect_ratio else intrinsic_dimensions

    default_alignment = cell.content_alignment

    # Compute the horizontal and vertical alignment factor from the intrinsic content dimensions and the focal point;
    # if no focal point is defined, use the alignment of the cell
    alignment = compute_focal_alignment(intrinsic_dimensions[0], intrinsic_dimensions[1], focal_point, default_alignment)

    # The default rendition dimensions are the intrinsic dimensions
    render_dimensions = intrinsic_dimensions
    adjustments = 0, 0, 0, 0

    if cell.content_fit is cell.FitCover:

        # Shrink the intrinsic dimensions to achieve the cell aspect ratio
        render_dimensions, adjustments = \
            apply_aspect_ratio(intrinsic_dimensions, cell_aspect_ratio, expand=False, alignment=alignment)

    # Copy any styles defined on the cell. They will be applied to the content element.

    styles = dict(cell.styles)

    if not is_replaced_element and cell.use_aspect_ratio and cell.content_fit is cell.FitContain:
        styles['--aspect-ratio'] = '{:d}/{:d}'.format(cell.aspect_ratio[0], cell.aspect_ratio[1])

    # Compute the setting for the object-position CSS property:

    # alignment = CSSUnitValue(value=alignment[0] * 100.0, unit=CSSUnits.PERCENTAGE_UNIT), \
    #             CSSUnitValue(value=alignment[1] * 100.0, unit=CSSUnits.PERCENTAGE_UNIT)

    # Express the content_fit and alignment settings as CSS properties
    # styles = apply_fit_and_alignment_styles(cell.content_fit, alignment, styles)

    extra_classes = []

    if cell.use_aspect_ratio and cell.content_fit is cell.FitContain:

        # We need to maintain the cell's aspect ratio if content_fit is contain.
        # A div wrapper with the desired aspect ratio allows us to scale the
        # the image appropriately.

        # extra_classes.append('fitted-aspect-ratio')
        pass

    # Format the element attributes

    attributes = dict(attributes) if attributes is not None else {}

    if styles:
        attributes['style'] = format_styles(styles)

    if cell.style_classes or extra_classes:
        attributes['class'] = cell.formatted_style_classes + " " + " ".join(extra_classes)

    sizes_list = []

    for condition, size in cell.conditional_sizes:
        sizes_list.append('({}) {}'.format(condition, size))

    sizes_list.append(str(cell.default_size))
    sizes_list = ', '.join(sizes_list)

    # Render the content
    result = renderer(cell, content, render_dimensions, adjustments, alignment, attributes, sizes_list)

    # We need to maintain the cell's aspect ratio if content_fit is contain and is_replaced_element is True.
    # A div wrapper with the desired aspect ratio is defined by using the padding "hack".

    if is_replaced_element and cell.use_aspect_ratio and cell.content_fit is cell.FitContain:

        # "position: relative; height: 0; padding-top: calc({:6f} * 100%);"

        result = '<div style="--aspect-ratio: {:d}/{:d};">{}</div>'.format(
                    cell.aspect_ratio[0], cell.aspect_ratio[1], result)

    return result



from django import template
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.templatetags.static import static

from ..cells import Cell, ResponsiveCell
from ..apps import get_app_label

APP_LABEL = get_app_label()

register = template.Library()


@register.simple_tag(name="render_responsive_cell", takes_context=True)
def render_responsive_cell_tag(context, responsive_delegate, cell, *, item=None):

    if not responsive_delegate:
        return ''

    if item is None:

        if cell.context is None:
            return ''

        item = cell.context.items[cell.source_order]

    result = responsive_delegate.render_in_responsive_cell(item, cell, template_context=context)
    result = mark_safe(result)
    return result


@register.simple_tag(name="define_cell")
def define_cell_tag(*,
                    render_index=0,
                    source_order=0,
                    default_size=None,
                    conditional_sizes=None,
                    aspect_ratio=None,
                    content_fit=None,
                    content_alignment=None,
                    styles=None,
                    style_classes=None,
                    annotations=None):

    result = Cell(render_index=render_index,
                  source_order=source_order,
                  default_size=default_size,
                  conditional_sizes=conditional_sizes,
                  aspect_ratio=aspect_ratio,
                  content_fit=content_fit,
                  content_alignment=content_alignment,
                  styles=styles,
                  style_classes=style_classes,
                  annotations=annotations)

    return result


@register.simple_tag(name="define_responsive_cell")
def define_responsive_cell_tag(responsive_context, *,
                               render_index=0,
                               source_order=0,
                               default_size=None,
                               conditional_sizes=None,
                               aspect_ratio=None,
                               content_fit=None,
                               content_alignment=None,
                               styles=None,
                               style_classes=None,
                               annotations=None):

    result = ResponsiveCell(render_index=render_index,
                            source_order=source_order,
                            default_size=default_size,
                            conditional_sizes=conditional_sizes,
                            aspect_ratio=aspect_ratio,
                            content_fit=content_fit,
                            content_alignment=content_alignment,
                            styles=styles,
                            style_classes=style_classes,
                            annotations=annotations)

    result.context = responsive_context
    return result


@register.simple_tag(takes_context=False)
def aldine_support(*, container_element):

    if container_element == 'head':
        return format_html('<link rel="stylesheet" type="text/css" href="{}">',
                           static(APP_LABEL + '/css/aldine.css'))

    return ''




import weakref
import types

from typing import Optional, Union

from steadfast import *

from csskit.common import CSSUnits, CSSUnitValue, CSSMathExpr, CSSObjectFitType, CSSObjectFitNone, \
    CSSObjectFitFill, CSSObjectFitCover, CSSObjectFitContain

from .auxiliaries import format_numeric_value, format_styles
from .apps import get_app_label
from .__dynamic_blocks_placeholder import Resolvable

__all__ = ['Cell', 'ResponsiveCell']

APP_LABEL = get_app_label()


class Cell:

    ContentFitType = CSSObjectFitType

    FitNone = CSSObjectFitNone
    FitContain = CSSObjectFitContain
    FitCover = CSSObjectFitCover
    FitFill = CSSObjectFitFill

    @property
    def render_index(self):
        return self.__render_index

    @property
    def source_order(self):
        return self.__source_order

    @property
    def default_size(self) -> Optional[Union[CSSUnitValue, CSSMathExpr]]:
        return self.__default_size

    @property
    def conditional_sizes(self):
        return self.__conditional_sizes[:] if isinstance(self.__conditional_sizes, list) else self.__conditional_sizes

    @property
    def use_aspect_ratio(self):
        return self.__aspect_ratio is not None

    @property
    def aspect_ratio(self):
        return self.__aspect_ratio

    @property
    def aspect_ratio_number(self):

        if not self.use_aspect_ratio:
            return None

        return self.__aspect_ratio[0] / self.__aspect_ratio[1]

    @property
    def aspect_ratio_as_text(self):

        if not self.use_aspect_ratio:
            return "1/1"

        return format_numeric_value(self.__aspect_ratio[0]) + "/" + \
               format_numeric_value(self.__aspect_ratio[1])

    @property
    def content_fit(self) -> ContentFitType:
        return self.__content_fit

    @property
    def content_alignment(self):
        return self.__content_alignment

    @property
    def styles(self):
        return self.__styles

    @property
    def formatted_styles(self):
        return format_styles(self.__styles)

    @property
    def style_classes(self):
        return self.__style_classes

    @property
    def formatted_style_classes(self):
        return " ".join(self.style_classes)

    @property
    def annotations(self):
        return self.__annotations

    @property
    def image_filter_expr(self):
        return self.__image_filter_expr

    def __init__(self, *,
                 render_index=0,
                 source_order=0,
                 default_size=None,
                 conditional_sizes=None,
                 aspect_ratio=None,
                 content_fit=None,
                 content_alignment=None,
                 styles=None,
                 style_classes=None,
                 annotations=None,
                 image_filter_expr=None):

        if conditional_sizes is None:
            conditional_sizes = []

        if isinstance(content_fit, str):
            content_fit = self.ContentFitType.from_identifier(content_fit)

        if content_fit is None:
            content_fit = self.FitCover

        if styles is None:
            styles = {}

        if style_classes is None:
            style_classes = []

        if annotations is None:
            annotations = {}

        if isinstance(annotations, dict):
            annotations = types.SimpleNamespace(**annotations)

        if not isinstance(annotations, types.SimpleNamespace):
            raise RuntimeError("annotations argument must be a dict or SimpleNamespace")

        self.__render_index = render_index
        self.__source_order = source_order

        self.__default_size = default_size
        self.__conditional_sizes = conditional_sizes \
            if isinstance(conditional_sizes, Resolvable) else \
            list(conditional_sizes)

        self.__aspect_ratio = aspect_ratio
        self.__content_fit = content_fit
        self.__content_alignment = content_alignment
        self.__styles = dict(styles)
        self.__style_classes = list(style_classes)

        self.__annotations = types.SimpleNamespace(**annotations.__dict__)
        self.__image_filter_expr = image_filter_expr


@decl_serializable(
    SaveInitArguments(),
    type_identifier=APP_LABEL + '.cell'  # noqa
)
class ResponsiveCell(Cell, Resolvable):

    @property
    def context(self):
        return self.__context_ref() if self.__context_ref else None

    @context.setter
    def context(self, value):
        self.__context_ref = weakref.ref(value) if value else None

    @property
    def media_conditions(self):
        return list(self.__definitions_by_media_condition.keys())

    def __init__(self, *,
                 render_index=0, source_order=0,
                 default_size=None, conditional_sizes=None,
                 aspect_ratio=None,
                 content_fit=None, content_alignment=None,
                 styles=None, style_classes=None,
                 annotations=None,
                 image_filter_expr=None,
                 cells_by_media_condition=None):

        Cell.__init__(self,
                      render_index=render_index, source_order=source_order,
                      default_size=default_size, conditional_sizes=conditional_sizes,
                      aspect_ratio=aspect_ratio,
                      content_fit=content_fit, content_alignment=content_alignment,
                      styles=styles,
                      style_classes=style_classes,
                      annotations=annotations,
                      image_filter_expr=image_filter_expr)

        Resolvable.__init__(self)

        if cells_by_media_condition is None:
            cells_by_media_condition = []

        self.__context_ref = None
        self.__cells_by_media_condition = list(cells_by_media_condition)
        self.__cells_dict = {}

        for condition, cell in cells_by_media_condition:
            self.__cells_dict[condition] = cell

    def save_init_arguments(self, arguments):

        arguments.render_index = self.render_index
        arguments.source_order = self.source_order
        arguments.default_size = self.default_size

        arguments.conditional_sizes = self.conditional_sizes \
            if isinstance(self.conditional_sizes, Resolvable) else \
            list(self.conditional_sizes)

        arguments.aspect_ratio = self.aspect_ratio
        arguments.content_fit = self.content_fit
        arguments.content_alignment = self.content_alignment
        arguments.styles = dict(self.styles)
        arguments.style_classes = list(self.style_classes)
        arguments.annotations = dict(self.annotations.__dict__)
        arguments.cells_by_media_condition = list(self.__cells_by_media_condition)

    def cell_for_media_condition(self, media_condition) -> Optional[Cell]:

        result = self.__cells_dict.get(media_condition, None)

        if result is None:
            return self

        return result

    def resolve(self, context, strict=False, cacheable=False, variables=None):

        args = types.SimpleNamespace()
        self.save_init_arguments(args)

        if isinstance(args.render_index, Resolvable):
            args.render_index = args.render_index.resolve(context, strict, cacheable, variables)

        if isinstance(args.source_order, Resolvable):
            args.source_order = args.source_order.resolve(context, strict, cacheable, variables)

        if isinstance(args.default_size, Resolvable):
            args.default_size = args.default_size.resolve(context, strict, cacheable, variables)

        if isinstance(args.conditional_sizes, Resolvable):
            args.conditional_sizes = args.conditional_sizes.resolve(context, strict, cacheable, variables)

        for index, item in enumerate(args.conditional_sizes):

            if isinstance(item, Resolvable):
                item = item.resolve(context, strict, cacheable, variables)
                args.conditional_sizes[index] = item
                continue

            condition, value = item

            if isinstance(value, Resolvable):
                value = value.resolve(context, strict, cacheable, variables)

            args.conditional_sizes[index] = condition, value

        if isinstance(args.aspect_ratio, Resolvable):
            args.aspect_ratio = args.aspect_ratio.resolve(context, strict, cacheable, variables)

        if isinstance(args.content_fit, Resolvable):
            args.content_fit = args.content_fit.resolve(context, strict, cacheable, variables)

        if isinstance(args.content_alignment, Resolvable):
            args.content_alignment = args.content_alignment.resolve(context, strict, cacheable, variables)

        for key, value in self.styles.items():

            if isinstance(value, Resolvable):
                value = value.resolve(context, strict, cacheable, variables)

            args.styles[key] = value

        for index, value in enumerate(args.style_classes):

            if isinstance(value, Resolvable):
                value = value.resolve(context, strict, cacheable, variables)

            args.style_classes[index] = value

        for index, cell in enumerate(args.cells_by_media_condition):
            cell = cell.resolve(context, strict, cacheable, variables)
            args.cells_by_media_condition[index] = cell

        for name, value in self.annotations.__dict__.items():

            if isinstance(value, Resolvable):
                value = value.resolve(context, None, cacheable, variables)

            args.annotations[name] = value

        args.image_filter_expr = self.image_filter_expr

        return self.__class__(**args.__dict__)




from types import SimpleNamespace

from wagtail_content_admin.blocks import BaseContentRenderBlock

from .apps import get_app_label
from .cells import ResponsiveCell

__all__ = ['BaseContentLayoutBlock']


APP_LABEL = get_app_label()


class BaseContentLayoutBlock(BaseContentRenderBlock):

    class Meta:

        icon = "placeholder"

        cells_and_items_var = "cells_and_items"

    # noinspection PyMethodMayBeStatic
    def create_cell_state(self, value, context):
        return SimpleNamespace()

    # noinspection PyMethodMayBeStatic
    def create_cell_for_item(self, item, render_index, source_index, annotations, cell_state):
        cell = ResponsiveCell(
            render_index=render_index,
            source_order=source_index,
            annotations=annotations
        )

        return cell

    def contribute_content_to_context(self, value, content, context):

        super(BaseContentLayoutBlock, self).contribute_content_to_context(value, content, context)

        cell_state = self.create_cell_state(value, context)
        cells_and_items = []

        for index, entry in enumerate(zip(content.items, content.annotations) if content.items else []):
            item, item_annotations = entry

            cell = self.create_cell_for_item(item, index, index, item_annotations, cell_state)
            cells_and_items.append((cell, item))

        if self.meta.cells_and_items_var:
            context[self.meta.cells_and_items_var] = cells_and_items



from django.urls import reverse
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class AldineConfig(AppConfig):
    name = 'aldine'
    label = 'aldine'
    verbose_name = _("Aldine")
    default_auto_field = 'django.db.models.BigAutoField'

    def ready(self):

        from .cells import ResponsiveCell

        pass


def get_app_label():
    return AldineConfig.label


def reverse_app_url(identifier):
    return reverse(f'{AldineConfig.label}:{identifier}')





__all__ = ['apply_aspect_ratio', 'clip_value', 'compute_focal_alignment', 'format_numeric_value',
           'format_styles', 'apply_fit_and_alignment_styles', 'format_attributes']


def clip_value(value, min_value, max_value):
    return max(min_value, min(value, max_value))


def compute_focal_alignment(width, height, focal_point, default_alignment=None):

    if focal_point:

        result = clip_value(focal_point[0] / width, 0.0, 1.0), \
                 clip_value(focal_point[1] / height, 0.0, 1.0)

    elif default_alignment:

        result = clip_value(default_alignment[0], 0.0, 1.0), \
                 clip_value(default_alignment[1], 0.0, 1.0)

    else:

        result = 0.5, 0.5

    return result


def apply_fit_and_alignment_styles(content_fit, content_alignment, styles=None):

    if styles is None:
        styles = {}

    if content_fit:
        styles['object-fit'] = content_fit.identifier

    if content_alignment:
        styles['object-position'] = "{} {}".format(content_alignment[0], content_alignment[1])

    return styles


def as_integer(operand):

    result = [0] * len(operand)

    for i in range(len(operand)):
        result[i] = int(round(operand[i]))

    result = tuple(result)
    return result


def apply_aspect_ratio(dimensions, aspect_ratio, expand=False, alignment=None, cast_to_integer=True):

    """
    Rescales the given dimensions to fit the provided aspect ratio. If expand is False,
    the dimensions will be shrunk along one axis, or expanded along one axis otherwise.
    The new dimensions will be returned together with a 4-tuple that indicates the change
    along each edge of a box defined by the old dimensions.

    :param dimensions: A 2-tuple of a width and a height
    :param aspect_ratio: A 2-tuple of a width and a height
    :param expand: If True, expand the dimensions along one axis, or shrink them along one axis if False.
    :param alignment: A numeric 2-tuple ([0..1], [0..1]) describing the relative alignment of the old dimensions relative to the new dimensions
    :param cast_to_integer: Cast all resulting values to integers.
    :return: A 4-tuple of top, right, bottom, left adjustments that when added to a box defined by
            the old dimensions produces a box defined by the new dimensions
    """

    x, y = aspect_ratio
    width, height = dimensions

    if alignment is None:
        alignment = 0.5, 0.5

    dx, dy = alignment
    dx, dy = clip_value(dx, 0.0, 1.0), clip_value(dy, 0.0, 1.0)

    aspect_ratio_number = x / y
    dimensions_ratio_number = width / height

    test = dimensions_ratio_number > aspect_ratio_number

    if test != expand:
        dimensions = aspect_ratio_number * height, height
        space = width - dimensions[0] if expand else dimensions[0] - width
        adjustments = 0, (1.0 - dx) * space, 0, dx * space
    elif dimensions_ratio_number == aspect_ratio_number:
        dimensions = width, height
        adjustments = 0, 0, 0, 0
    else:
        dimensions = width, width / aspect_ratio_number
        space = height - dimensions[1] if expand else dimensions[1] - height
        adjustments = dy * space, 0, (1.0 - dy) * space, 0

    if cast_to_integer:
        dimensions = as_integer(dimensions)
        adjustments = as_integer(adjustments)

    return dimensions, adjustments


def format_numeric_value(value, decimals=3):

    try:
        value = '{:d}'.format(value)
    except ValueError:
        f = '{{:.{:d}f}}'.format(decimals)
        value = f.format(value)

    return value


def format_styles(styles):

    result = []

    for key, value in styles.items():
        result.append("{}: {};".format(key, value))

    return " ".join(result)


def format_attributes(attributes):

    if not attributes:
        attributes = ''
    else:
        attributes = [name + '="' + value + '"' for name, value in attributes.items()]
        attributes = " ".join(attributes)

    return attributes



import bisect

from types import SimpleNamespace
from urllib.parse import quote

from .auxiliaries import as_integer, format_attributes, apply_fit_and_alignment_styles, format_styles

from .render import render_content_in_responsive_cell

from csskit.common import CSSUnits, CSSUnitValue

__all__ = ['ImageCurator', 'render_image_in_responsive_cell']

LANDSCAPE_RESOLUTIONS = [240, 320, 410, 640, 800, 1024, 1366, 1600, 2560, 3840, 5120, 7680]
PORTRAIT_RESOLUTIONS = [320, 480, 768, 900, 1200, 1440, 2160, 2880, 4320]


class ImageCurator(object):

    def __init__(self):
        self.__resolutions = SimpleNamespace()
        self.__resolutions.landscape = LANDSCAPE_RESOLUTIONS
        self.__resolutions.portrait = PORTRAIT_RESOLUTIONS

    def select_resolutions(self, dimensions):

        width, height = dimensions

        if width > height:
            resolutions = self.__resolutions.landscape
            value = width
            other_value = height

            def scaled_dimensions(value, other_value):
                return value, other_value

        else:
            resolutions = self.__resolutions.portrait
            value = height
            other_value = width

            def scaled_dimensions(value, other_value):
                return other_value, value

        index = bisect.bisect_left(resolutions, value)
        selected_resolutions = resolutions[:index]

        result = []

        for scaled_value in selected_resolutions:
            scaled_other_value = other_value * scaled_value / value
            result.append(scaled_dimensions(scaled_value, scaled_other_value))

        result.append(scaled_dimensions(value, other_value))
        return result


def render_img_tag(cell, content, dimensions, adjustments, alignment, attributes, layout_sizes, *, curator):

    image = content

    alignment = CSSUnitValue(value=alignment[0] * 100.0, unit=CSSUnits.PERCENTAGE_UNIT), \
                CSSUnitValue(value=alignment[1] * 100.0, unit=CSSUnits.PERCENTAGE_UNIT)

    # Express the content_fit and alignment settings as CSS properties
    alignment_styles = apply_fit_and_alignment_styles(cell.content_fit, alignment)
    alignment_styles = format_styles(alignment_styles)

    styles = attributes.get('style', '')

    if styles:
        styles += ' '

    attributes['style'] = styles + alignment_styles

    attributes = format_attributes(attributes)

    # Compute the rendition sizes based on the (computed) rendition dimensions
    rendition_sizes = [as_integer(size) for size in curator.select_resolutions(dimensions)]

    # Render the content in the selected rendition sizes

    renditions = []
    source_set = []

    for index, size in enumerate(rendition_sizes):
        rendition_filter = 'fill-{:d}x{:d}'.format(size[0], size[1])

        if cell.image_filter_expr:
            rendition_filter += "|" + cell.image_filter_expr

        rendition = image.get_rendition(rendition_filter)
        source_set.append('{} {:d}w'.format(quote(rendition.url), size[0]))
        renditions.append(rendition)

    base_url = renditions[-1].url

    result = '<img srcset="{}" sizes="{}" src="{}" alt="{}"{}>'.format(
        ', '.join(source_set), layout_sizes, quote(base_url), image.default_alt_text,
        attributes
    )

    return result


def render_image_in_responsive_cell(image, cell, curator, attributes=None):

    image_dimensions = image.width, image.height

    return render_content_in_responsive_cell(
        image, image_dimensions, image.get_focal_point(), cell,
        lambda *args: render_img_tag(*args, curator=curator),
        attributes=attributes)



import copy

from django.utils.deconstruct import deconstructible
from steadfast import *

__all__ = ['Resolvable', 'ResponsiveDelegate']

SERIALISATION_PREFIX = "wagtail_dynamic_blocks."


@decl_serializable(
    SaveInitArguments(),
    type_identifier=SERIALISATION_PREFIX + 'resolvable'  # noqa
)
class Resolvable(object):

    def __init__(self):
        pass

    def save_init_arguments(self, arguments):
        pass

    def copy(self):
        return copy.deepcopy(self)

    def resolve(self, context, strict=False, cacheable=False, variables=None):
        return self


@deconstructible
class ResponsiveDelegate:

    def render_in_responsive_cell(self, item, cell, template_context=None):
        return ''
